// equivtest.cc,
// program to test functional equivalence of forint calculation modes

#include <iostream>
#include <math.h>
#include <stdio.h>

#include <forint/forint.hh>

using namespace FInt;

double alpha=1., beta=1.2, var1=2, var2=-1;
double  A [2][3];
double* C; 

double f0(void)   { return M_PI; }

double f1(double a)
{
  if (a < 0) return a;
  if (a == 0) return 1.;
  return (long long)a * f1((long long)(a-1));
}

double f2(double a, double b)
{ return (a) ? a * b : b; }

double f3(void* a, void* b)
{ return ((double*)a)[0] * ((double*)b)[1]; }

double f4(double i, void* a)
{ return (i > 7 && i < 0) ? 0. : ((double*)a) [long(i)]; }

void err_quit(void) { cerr << "\n-----ERROR QUIT-----" << endl; exit(1); }

main()
{
  mssg cli;
  forint calc(cli);

  if (calc.putConst(alpha,"alpha"))
    err_quit();
  if (calc.putVar(beta, "beta"))
    err_quit();
  if (calc.putVar(var1, "var1"))
    err_quit();
  if (calc.putVar(var2, "var2"))
    err_quit();

  for (int i=0; i<2; i++)
    for (int j=0; j<3; j++)
      A[i][j] = i*3+j;
  vector<int> dims_A(2); 
  dims_A[0] = 2; 
  dims_A[1] = 3; 
  if (calc.putPoin(A, "A", "double", &dims_A))
    err_quit();

  C = new double[5];
  for (int i=0; i<5; i++)
    C[i] = 10+i;
  vector<int> dims_C(1);
  dims_C[0] = 5;
  if (calc.putPoin(C, "C", "double", &dims_C))
    err_quit();

  if (calc.putFunc0 (f0, "f0"))
    err_quit();
  if (calc.putFunc1 (f1, "f1"))
    err_quit();
  if (calc.putFunc2D(f2, "f2"))
    err_quit();
  if (calc.putFunc2P(f3, "f3","double*","double*"))
    err_quit();
  if (calc.putFunc2 (f4, "f4","double*"))
    err_quit();

// test expression
  /* char* expr = "f4(C,var2)";*/
  char* expr = "(sin(alpha) > 0.5 && beta <= 1)
      ? var1 - sqrt(var2/10)+f4(2,A[0])+C[4-var2]
      : f0()+f1(var2)+f2(var1,var2)+f3(A[1],C)";
  /*char* expr = "var1 + f4(2,A[0])-C[4-var2]";*/
  /*char* expr = "f4(A[0],2)";*/

  cout << "\nProgram to test functional equivalence of forint "
    "calculation modes:\n\n  RVM_MODE, SVM_MODE, DLF_MODE.\n\n"
    "expression uses external:\n"
    "  constant:  double alpha;\n"
    "  variable:  double beta;\n"
    "  pointers:  double A[2][3], **C,\n"
    "  function:  double f0(void);\n"
    "             double f1(double);\n"
    "             double f2(double, double);\n"
    "             double f3(double*, double*);\n"
    "             double f4(double, double*);\n\n"
    "A testing expression:\n\n"
    "  \'" << expr << "\'\n\n"
    "Press <Enter> to start (you have to wait for a DLF generating)..."
       << flush;
  getchar();

  cout << "\nParsing..." << flush;
  if (calc.parse(expr)) {
    cerr << endl << "parse(), error exiting\n";
    exit(1);
  }
  cout << "done" << endl;
  
  cout << "Generating of a dynamically linked function..." << flush;
  if (calc.makeDLF()) {
    cerr << endl << "makeDLF(), error exiting\n";
    getchar();
    exit(1);
  }
  cout << "done" << endl;
  
  volatile double value = 0;
  cout << "A value is calculated by the formula above in modes above\n"
    "Press <Enter> to start..." << flush;
  getchar();
  cout << "\n beta var2   "
       << "RVM         SVM         DLF         built-in formula"
       << endl;
  for (int i=0; i<5; i++) {
    beta -= 0.1;
    var2 += 1;
    value = calc.calculate(RVM_MODE);
    printf(" %.2f %.2f   %.5e",beta,var2,value);
    value = calc.calculate(SVM_MODE);
    printf(" %.5e",value);
    value = calc.calculate(DLF_MODE);
    printf(" %.5e",value);
    /*value = f4(A[0],2);*/
    value = (sin(alpha) > 0.5 && beta <= 1)
      ? var1 - sqrt(var2/10)+f4(2,A[0])+C[4-long(var2)]
      : f0()+f1(var2)+f2(var1,var2)+f3(A[1],C);
    printf(" %.5e\n",value);
  }
}









